/*
  This program is free software: you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation, either version 3 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

/* 
   Experiments with the SainSmart DDS120 USB scope

   Heavely based on donut6's scope.cpp, see: 
   http://www.eevblog.com/forum/testgear/sainsmart-dds120-usb-oscilloscope-(buudai-bm102)/msg536505/#msg536505
   and doctormord's analysis of the DDS120 USB communication protocol, see: 
   http://www.eevblog.com/forum/testgear/sainsmart-dds120-usb-oscilloscope-(buudai-bm102)/msg534769/#msg534769 
   and 
   http://www.360customs.de/en/2014/10/usb-oszilloskop-sainsmart-dds120-2-kanal-20mhz-50msps-buudairocktech-bm102

   mmark
*/


#include <errno.h>
#include <signal.h>
#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <string>

#include <libusb-1.0/libusb.h>
#include <unistd.h>

#define BUF_SIZE 0x2000

struct libusb_device_handle *scope_handle;
struct libusb_transfer *irq_transfer;

unsigned char scope_data[BUF_SIZE];
int do_exit = 0;
bool gotIrq = false;

int scope_request(int req, int val)
{
    unsigned char data = 0;
    if (scope_handle)
        int r = libusb_control_transfer(scope_handle, 0 | LIBUSB_ENDPOINT_IN, req, val, 0, &data, 1, 0);
    printf("request: 0x%02x,0x%02x: 0x%02x\n", req, val, data);
    usleep(1000);
    return data;
}


void init_dds120()
{
    scope_request(0x24, 0x08); // Ch1 coupling 0x08:DC, 0x00:AC
    scope_request(0x25, 0x01); // Ch2 coupling 0x01:DC, 0x00:AC
    scope_request(0x94, 0x10); // Sampling rate 0x10:48MHz, 0x01:2.4MHz, 0x11:240kHz 
    scope_request(0x22, 0x08); // Ch1 sensitivity (/div): 0x08:50mV, 0x04:100mV, 0x00:200mV, 0x06:500mV, 0x02:1-5V
    scope_request(0x23, 0x20); // Ch2 sensitivity (/div): 0x20:50mV, 0x10:100mV, 0x00:200mV, 0x12:500mV, 0x02:1-5V

//#define INIT90
#ifdef INIT90
    scope_request(0x90, 0x01);
    scope_request(0x90, 0x02);
    scope_request(0x90, 0x05);
    scope_request(0x90, 0x03);
    scope_request(0x90, 0x06);
    scope_request(0x90, 0x04);
    scope_request(0x90, 0x07);
    scope_request(0x90, 0x08);
    scope_request(0x90, 0x09);
    scope_request(0x90, 0x0A);
    scope_request(0x90, 0x0B);
    scope_request(0x90, 0x0C);
    scope_request(0x90, 0x0D);
    scope_request(0x90, 0x0E);
    scope_request(0x90, 0x0F);
    scope_request(0x90, 0x10);
    scope_request(0x90, 0x11);
    scope_request(0x90, 0x12);
    scope_request(0x90, 0x13);
#endif

}


int  scope_get_buffer(unsigned char* buf, size_t buf_size)
{
    int r;
    int transferred = 0;
    if (scope_handle) {
        r = libusb_bulk_transfer(scope_handle, 0x2 | LIBUSB_ENDPOINT_IN, buf, buf_size, &transferred, 500);
        if (r < 0) {
            fprintf(stderr, "scope data read error %d\n", r);
            return r;
        }
    }
    return transferred;
}


int scope_acquire()
{
    int w = 0;
    int resync = 0;
    printf("scope_acquire\n");
    scope_request(0x33, 0x00);   // "go" code

    printf("check fifo status...\n");
    // check FIFO status:
    while ((w != 0x2c) && (w != 0x21)) {
        w = scope_request(0x50, 0);
        resync++;
        if (resync > 1000) 
            break;
    }
    printf("w: %d, resync: %d\n", w, resync);

    while (w != 0x21) {
        w = scope_request(0x50, 0);
        resync++;
        if (resync > 2000) 
            break;
    }
    printf("w: %d, resync: %d\n", w, resync);

    if (w == 0x21) { // fifo full
        printf("fifo full, getting buffer...\n");
        int r = scope_get_buffer(scope_data, sizeof(scope_data));
        printf("got %d bytes\n", r);
        if (r < 0)
            return 0;
        return r;
    } else
        printf("no data received\n");

    return 0;
}


void render_data(int n)
{
    // display ch1 and ch2...
    for (int i = 0; i < (n-1); i += 2) {
        int y1 = scope_data[i] - 0x80;
        int y2 = scope_data[i + 1] - 0x80;
        printf("%6d %6d %6d\n", i, y1, y2);
    }
}

static void cb_irq(struct libusb_transfer *transfer)
{
    if (transfer->status != LIBUSB_TRANSFER_COMPLETED) {
        fprintf(stderr, "irq transfer status %d?\n",
                transfer->status);
        do_exit = 2; 
        libusb_free_transfer(transfer);
        irq_transfer = NULL;
        return;
    }
    gotIrq = true;
    printf("IRQ callback got %d bytes\n", transfer->actual_length);
}

static int alloc_transfers(void)
{
    irq_transfer = libusb_alloc_transfer(0);

    if (!irq_transfer)
        return -ENOMEM;

    libusb_fill_interrupt_transfer(irq_transfer, scope_handle, 0x2 | LIBUSB_ENDPOINT_IN,
                                   scope_data, sizeof(scope_data), cb_irq, NULL, 0);

    return 0;
}


int main(int argc, char *argv[])
{
    int r = libusb_init(NULL);
    if (r < 0) {
        fprintf(stderr, "ERROR: Can't initialize libusb (%d)\n", r);
        return -1;
    }

	printf("looking for device 8102:8102\n");
	scope_handle = libusb_open_device_with_vid_pid(NULL, 0x8102, 0x8102);
	if (scope_handle == 0) {
	    fprintf(stderr, "Could not open device\n");
	    return -1;
	} else
	    printf("found a DDS120!\n");

	r = libusb_claim_interface(scope_handle, 0);
	if (r < 0) {
	    fprintf(stderr, "libusb_claim_interface error %d\n", r);
	    return -1;
	}

    init_dds120();

#define USE_SYNC_TRANSFER
#ifdef USE_SYNC_TRANSFER

    while (true) {
        int r = scope_get_buffer(scope_data, sizeof(scope_data));
        printf("#got %d bytes\n", r);
        render_data(r);
    }

#else // asynchronous transfer

    r = alloc_transfers();
    if (r < 0) {
	    fprintf(stderr, "libusb_alloc_transfer error %d\n", r);
	    return -1;
	}

    while (!do_exit) {
        scope_request(0x33, 0x00);   // "go" code
        r = libusb_submit_transfer(irq_transfer);
        if (r < 0) {
            fprintf(stderr, " libusb_submit_transfer error %d\n", r);
            return -1;
        }
        libusb_handle_events(NULL);
        if (gotIrq) {
            render_data(irq_transfer->actual_length);
            gotIrq = false;
        }
        usleep(10);
    }
#endif

	return 0;
}
